/*------------------------------------------------------------------------------
* nvs.c : NVS receiver dependent functions
*
*          Copyright (C) 2012 by M.BAVARO and T.TAKASU, All rights reserved.
*
* reference :
*     [1] Description of BINR messages which is used by RC program for RINEX
*         files accumulation, NVS
*     [2] NAVIS Navis Standard Interface Protocol BINR, NVS
*
* version : $Revision:$ $Date:$
* history : 2012/01/30 1.0  first version by M.BAVARO
*           2012/04/05 1.1  modified by T.TAKASU
*-----------------------------------------------------------------------------*/
#include "rtklib.h"

#define NVSSYNC     0x10      /* nvs message sync code 1 */
#define NVSENDMSG   0x03      /* nvs message sync code 1 */
#define NVSCFG      0x06      /* nvs message cfg-??? */

#define ID_XF5RAW   0xf5      /* nvs msg id: raw measurement data */
#define ID_X4AIONO  0x4a      /* nvs msg id: gps ionospheric data */
#define ID_X4BTIME  0x4b      /* nvs msg id: GPS/GLONASS/UTC timescale data */
#define ID_XF7EPH   0xf7      /* nvs msg id: subframe buffer */

static const char rcsid[]="$Id: nvs.c,v 1.0 2012/01/30 00:05:05 MBAVA Exp $";

/* get/set fields (little-endian) --------------------------------------------*/
#define U1(p)       (*((unsigned char  *)(p)))
#define U2(p)       (*((unsigned short *)(p)))
#define U4(p)       (*((unsigned int   *)(p)))
#define I1(p)       (*((char   *)(p)))
#define I2(p)       (*((short  *)(p)))
#define I4(p)       (*((int    *)(p)))
#define R4(p)       (*((float  *)(p)))

static double R8(const unsigned char *p)
{
    double value;
    unsigned char *q=(unsigned char *)&value;
    int i;
    for (i=0;i<8;i++) *q++=*p++;
    return value;
}

static const double ura_eph[]={
    2.4,3.4,4.85,6.85,9.65,13.65,24.0,48.0,96.0,192.0,384.0,768.0,1536.0,
    3072.0,6144.0,0.0
};
/* ura value (m) to ura index ------------------------------------------------*/
static int uraindex(double value)
{
    int i;
    for (i=0;i<15;i++) if (ura_eph[i]>=value) break;
    return i;
}
/* decode NVS xf5-raw: raw measurement data ----------------------------------*/
static int decode_xf5raw(raw_t *raw)
{
    gtime_t time;
    double tow,gpst_utc,glot_utc,cp1,pr1,dop1,tadj=0.01,toff=0.0,tn,freq;
    int i,j,n=0,nsat,week,sys,prn,frqn,snr,sat=0,flag,rcvtcorr;
    unsigned char *p=raw->buff+2;
    
    trace(4,"decode_xf5raw: len=%d\n",raw->len);
    
    /* no time tag adjustment option (-NOTADJ) */
    if (strstr(raw->opt,"-NOTADJ")) {
        tadj=0.0;
    }
    tow=R8(p);
    week=adjgpsweek(U2(p+8));
    gpst_utc=R8(p+10);
    glot_utc=R8(p+18);
    rcvtcorr=I1(p+26);
    nsat=(raw->len-29)/30;
    
    /* time in gpst */
    time=gpst2time(week,(tow+gpst_utc+rcvtcorr)*0.001);
    
    /* time tag adjustment */
    if (tadj>0.0) {
        tn=time2gpst(time,&week)/tadj;
        toff=(tn-floor(tn+0.5))*tadj;
        time=timeadd(time,-toff);
    }
    if (fabs(timediff(time,raw->time))<=0.001) {
        trace(2,"nvs xf5raw time tag duplicated: time=%s\n",time_str(time,3));
        return 0;
    }
    for (i=0,p+=27;i<nsat&&i<MAXOBS;i++,p+=30) {
        
        switch (U1(p)) {
            case 1: sys=SYS_GLO; break;
            case 2: sys=SYS_GPS; break;
            case 4: sys=SYS_SBS; break;
            case 8: sys=SYS_GAL; break;
            default:
                trace(2,"nvs xf5-raw system error: sys=%d\n",U1(p));
                continue;
        }
        prn =U1(p+ 1);
        frqn=I1(p+ 2);
        snr =I1(p+ 3);
        cp1 =R8(p+ 4);
        pr1 =R8(p+12)*0.001*CLIGHT;
        dop1=R8(p+20);
        flag=I1(p+28);
        
        trace(4,"sat=%2d L=%13.3f P=%13.3f D=%9.3f flags=%02X\n",sat,cp1,pr1,
              dop1,flag);
        
        if (!(sat=satno(sys,prn))) {
            trace(2,"nvs xf5-raw satellite number error: sys=%d prn=%d\n",sys,prn);
            continue;
        }
        freq=sys==SYS_GLO?FREQ1_GLO+frqn*DFRQ1_GLO:FREQ1;
        raw->obs.data[n].time=time;
        raw->obs.data[n].sat=sat;
        raw->obs.data[n].L[0]=cp1-toff*freq;
        raw->obs.data[n].P[0]=pr1-toff*CLIGHT;
        raw->obs.data[n].D[0]=(float)dop1;
        raw->obs.data[n].SNR[0]=(unsigned char)(snr*4.0+0.5);
        raw->obs.data[n].LLI[0]=((flag&0x04)&&(flag&0x20))?0:1;
        raw->obs.data[n].code[0]=CODE_L1C;
        
        for (j=1;j<NFREQ;j++) {
            raw->obs.data[n].L[j]=raw->obs.data[n].P[j]=0.0;
            raw->obs.data[n].D[j]=0.0;
            raw->obs.data[n].SNR[j]=raw->obs.data[n].LLI[j]=0;
            raw->obs.data[n].code[j]=CODE_NONE;
        }
        n++;
    }
    raw->time=time;
    raw->obs.n=n;
    return 1;
}
/* decode ephemeris ----------------------------------------------------------*/
static int decode_gpsephem(int sat, raw_t *raw)
{
    eph_t eph={0};
    unsigned char *p=raw->buff+2;
    int week;
    double toc;
    
    trace(4,"decode_ephem: sat=%2d\n",sat);
    
    eph.crs   =R4(p+  2);
    eph.deln  =R4(p+  6)*1e+3;
    eph.M0    =R8(p+ 10);
    eph.cuc   =R4(p+ 18);
    eph.e     =R8(p+ 22);
    eph.cus   =R4(p+ 30);
    eph.A     =pow(R8(p+34),2);
    eph.toes  =R8(p+ 42)*1e-3;
    eph.cic   =R4(p+ 50);
    eph.OMG0  =R8(p+ 54);
    eph.cis   =R4(p+ 62);
    eph.i0    =R8(p+ 66);
    eph.crc   =R4(p+ 74);
    eph.omg   =R8(p+ 78);
    eph.OMGd  =R8(p+ 86)*1e+3;
    eph.idot  =R8(p+ 94)*1e+3;
    eph.tgd[0]=R4(p+102)*1e-3;
    toc       =R8(p+106)*1e-3;
    eph.f2    =R4(p+114)*1e+3;
    eph.f1    =R4(p+118);
    eph.f0    =R4(p+122)*1e-3;
    eph.sva   =uraindex(I2(p+126));
    eph.iode  =I2(p+128);
    eph.iodc  =I2(p+130);
    eph.code  =I2(p+132);
    eph.flag  =I2(p+134);
    week      =I2(p+136);
    
    eph.fit   =0; /* FIXME, there is no FIT time in the ephemeris? */
    
    eph.week=adjgpsweek(week);
    eph.toe=gpst2time(eph.week,eph.toes);
    eph.toc=gpst2time(eph.week,toc);
    eph.ttr=raw->time;
    
    if (!strstr(raw->opt,"-EPHALL")) {
        if (eph.iode==raw->nav.eph[sat-1].iode) return 0; /* unchanged */
    }
    eph.sat=sat;
    raw->nav.eph[sat-1]=eph;
    raw->ephsat=sat;
    return 2;
}
/* adjust daily rollover of time ---------------------------------------------*/
static gtime_t adjday(gtime_t time, double tod)
{
    double ep[6],tod_p;
    time2epoch(time,ep);
    tod_p=ep[3]*3600.0+ep[4]*60.0+ep[5];
    if      (tod<tod_p-43200.0) tod+=86400.0;
    else if (tod>tod_p+43200.0) tod-=86400.0;
    ep[3]=ep[4]=ep[5]=0.0;
    return timeadd(epoch2time(ep),tod);
}
/* decode glonass ephemeris --------------------------------------------------*/
static int decode_gloephem(int sat, raw_t *raw)
{
    geph_t geph={0};
    unsigned char *p=raw->buff+2;
    int prn,tk,tb;
    
    if (raw->len<93) {
        trace(2,"nvs NE length error: len=%d\n",raw->len);
        return -1;
    }
    prn        =I1(p+ 1);
    geph.frq   =I1(p+ 2);
    geph.pos[0]=R8(p+ 3);
    geph.pos[1]=R8(p+11);
    geph.pos[2]=R8(p+19);
    geph.vel[0]=R8(p+27)*1e+3;
    geph.vel[1]=R8(p+35)*1e+3;
    geph.vel[2]=R8(p+43)*1e+3;
    geph.acc[0]=R8(p+51)*1e+6;
    geph.acc[1]=R8(p+59)*1e+6;
    geph.acc[2]=R8(p+67)*1e+6;
    tb = (int)(R8(p+75)*1e-3);
    tk = tb;
    geph.gamn  =R4(p+83);
    geph.taun  =R4(p+87)*1e-3;
    geph.age   =I2(p+91);
    
    if (!(geph.sat=satno(SYS_GLO,prn))) {
        trace(2,"nvs NE satellite error: prn=%d\n",prn);
        return -1;
    }
    if (raw->time.time==0) return 0;
    
    geph.iode=(tb/900)&0x3F;
    geph.toe=utc2gpst(adjday(raw->time,tb-10800.0));
    geph.tof=utc2gpst(adjday(raw->time,tk-10800.0));
    
    raw->nav.geph[prn-1]=geph;
    raw->ephsat=geph.sat;
    return 2;
}
/* decode NVS ephemeris ----------------------------------------------------*/
static int decode_xf7eph(raw_t *raw)
{
    int prn,sat,sys;
    unsigned char *p=raw->buff;
    
    trace(4,"decode_xf7eph: len=%d\n",raw->len);
    
    if (raw->len<93) {
        trace(2,"nvs xf7eph length error: len=%d\n",raw->len);
        return -1;
    }
    sys = (U1(p+2)==1)?SYS_GPS:((U1(p+2)==2)?SYS_GLO:SYS_NONE);
    prn = U1(p+3);
    if (!(sat=satno(sys,prn))) {
        trace(2,"nvs xf7eph satellite number error: sys=%d prn=%d\n",sys,prn);
        return -1;
    }
    if (sys==SYS_GPS) {
        return decode_gpsephem(sat,raw);
    }
    else if (sys==SYS_GLO) {
        return decode_gloephem(sat,raw);
    }
    return 0;
}
/* decode NVS iono parameters ----------------------------------------------*/
static int decode_x4aiono(raw_t *raw)
{
    unsigned char *p=raw->buff+2;
    
    trace(4,"decode_x4aiono: len=%d\n", raw->len);
    
    raw->nav.ion_gps[0]=R4(p   );
    raw->nav.ion_gps[1]=R4(p+ 4);
    raw->nav.ion_gps[2]=R4(p+ 8);
    raw->nav.ion_gps[3]=R4(p+12);
    raw->nav.ion_gps[4]=R4(p+16);
    raw->nav.ion_gps[5]=R4(p+20);
    raw->nav.ion_gps[6]=R4(p+24);
    raw->nav.ion_gps[7]=R4(p+28);
    return 0;
}
/* decode NVS utc parameters -----------------------------------------------*/
static int decode_x4btime(raw_t *raw)
{
    unsigned char *p=raw->buff+2;
    
    trace(4,"decode_x4btime: len=%d\n", raw->len);
    
    raw->nav.utc_gps[1]=R8(p   );
    raw->nav.utc_gps[0]=R8(p+ 8);
    raw->nav.utc_gps[2]=I4(p+16);
    raw->nav.utc_gps[3]=I2(p+20);
    raw->nav.leaps     =I1(p+22);
    return 0;
}
/* decode NVS raw message --------------------------------------------------*/
static int decode_nvs(raw_t *raw)
{
    int type=U1(raw->buff+1);
    
    trace(0,"decode_nvs: type=%02x len=%d\n",type,raw->len);
    
    if (raw->outtype) {
        sprintf(raw->msgtype,"NVS 0x%02X (%4d):",type,raw->len);
    }
    switch (type) {
        case ID_XF5RAW : return decode_xf5raw (raw);
        case ID_XF7EPH : return decode_xf7eph (raw);
        case ID_X4AIONO: return decode_x4aiono(raw);
        case ID_X4BTIME: return decode_x4btime(raw);
        default: break;
    }
    return 0;
}
/* input NVS raw message from stream -----------------------------------------
* fetch next NVS raw data and input a message from stream
* args   : raw_t *raw   IO     receiver raw data control struct
*            raw->opt : nvs raw options
*                "-EPHALL"  : output all ephemerides
*          unsigned char data I stream data (1 byte)
* return : status (-1: error message, 0: no message, 1: input observation data,
*                  2: input ephemeris, 3: input sbas message,
*                  9: input ion/utc parameter)
*-----------------------------------------------------------------------------*/
extern int input_nvs(raw_t *raw, unsigned char data)
{
    trace(5,"input_nvs: data=%02x\n",data);
    
    /* synchronize frame */
    if (raw->nbyte==0&&data==NVSSYNC) {
        raw->buff[0]=data;
        raw->nbyte++;
        return 0;
    }
    if (raw->nbyte==1&&data!=NVSSYNC&&data!=NVSENDMSG) {
        
        /* Discard double 0x10 and 0x10 0x03 at beginning of frame */
        raw->buff[1]=data;
        raw->nbyte++;
        raw->flag=0;
        return 0;
    }
    /* This is all done to discard a double 0x10 */
    if (data==NVSSYNC) raw->flag=(raw->flag+1)%2;
    
    /* Store the new byte */
    if (data!=NVSSYNC||raw->flag) {
        raw->buff[raw->nbyte++]=data;
    }
    /* Detect ending sequence */
    if (data==NVSENDMSG&&raw->flag) {
        raw->len=raw->nbyte;
        raw->nbyte=0;
        
        /* Decode NVS raw message */
        return decode_nvs(raw);
    }
    if (raw->nbyte==MAXRAWLEN) {
        trace(2,"nvs message size error: len=%d\n",raw->nbyte);
        raw->nbyte=0;
        return -1;
    }
    return 0;
}
/* input NVS raw message from file -------------------------------------------
* fetch next NVS raw data and input a message from file
* args   : raw_t  *raw   IO     receiver raw data control struct
*          FILE   *fp    I      file pointer
* return : status(-2: end of file, -1...9: same as above)
*-----------------------------------------------------------------------------*/
extern int input_nvsf(raw_t *raw, FILE *fp)
{
    int i,data, odd=0;
    
    trace(4,"input_nvsf:\n");
    
    /* synchronize frame */
    for (i=0;;i++) {
        if ((data=fgetc(fp))==EOF) return -2;
        
        /* Search a 0x10 */
        if (data==NVSSYNC) {
            /* Store the frame begin */
            raw->buff[0] = data;
            if ((data=fgetc(fp))==EOF) return -2;
            
            /* Discard double 0x10 and 0x10 0x03 */
            if ((data != NVSSYNC) &&
                (data != NVSENDMSG)) {
                raw->buff[1]=data;
                break;
            }
        }
        if (i>=4096) return 0;
    }
    raw->nbyte = 2;
    for (i=0;;i++) {
        if ((data=fgetc(fp))==EOF) return -2;
        if (data==NVSSYNC) odd=(odd+1)%2;
        if ((data!=NVSSYNC) || odd) {
            /* Store the new byte */
            raw->buff[(raw->nbyte++)] = data;
        }
        /* Detect ending sequence */
        if ((data==NVSENDMSG) && odd) break;
        if (i>=4096) return 0;
    }
    raw->len = i;
    if ((raw->len) > MAXRAWLEN) {
        trace(2,"nvs length error: len=%d\n",raw->len);
        return -1;
    }
    /* decode nvs raw message */
    return decode_nvs(raw);
}
/* generate NVS binary message ---------------------------------------------
* generate NVS binary message from message string
* args   : char  *msg   I      message string
*            "RESET    [opt]" system reset
*            "WORK-SNS [sys]" work SNS setting
*            "DISABLE"        disable transmission
*          unsigned char *buff O binary message
* return : length of binary message (0: error)
*-----------------------------------------------------------------------------*/
extern int gen_nvs(const char *msg, unsigned char *buff)
{
    unsigned char *q=buff;
    char mbuff[1024],*args[32],*p;
    int narg=0;
    
    trace(4,"gen_nvs: msg=%s\n",msg);
    
    strcpy(mbuff,msg);
    for (p=strtok(mbuff," ");p&&narg<32;p=strtok(NULL," ")) {
        args[narg++]=p;
    }
    *q++=NVSSYNC;
    
    if (!strcmp(args[0],"RESET")) { /* reset/restart */
        /* !NVS RESET [opt] */
        *q++=0x01;
        *q++=0x00;
        *q++=0x01;
        *q++=0x21;
        *q++=0x01;
        *q++=0x00;
        *q++=narg>1?(unsigned char)atoi(args[1]):1; /* 0: with almanac delete */
    }
    else if (!strcmp(args[0],"WORK-SNS")) { /* work SNS setting */
        /* !NVS WORK-SNS [sys] */
        /* sys 0:GPS+GLO,1:GPS,2:GLO,10:GPS+GLO+SBS,11:GPS+SBS,12:reserved */
        *q++=0x0D;
        *q++=0x02;
        *q++=narg>1?(unsigned char)atoi(args[1]):0;
    }
    else if (!strcmp(args[0],"DISABLE")) { /* disable transmission */
        *q++=0x0E;
    }
    else if (!strcmp(args[0],"ENA-GPS")) { /* enable gps satellite */
        *q++=0x12;
        *q++=0x01;
        *q++=narg>1?(unsigned char)atoi(args[1]):1; /* prn */
        *q++=0x01;
    }
    else if (!strcmp(args[0],"ENA-GLO")) { /* enable glo satellite */
        *q++=0x12;
        *q++=0x02;
        *q++=narg>1?(unsigned char)atoi(args[1]):1; /* slot */
        *q++=0x01;
    }
    else if (!strcmp(args[0],"DIS-GPS")) { /* disable gps satellite */
        *q++=0x12;
        *q++=0x01;
        *q++=narg>1?(unsigned char)atoi(args[1]):1; /* prn */
        *q++=0x02;
    }
    else if (!strcmp(args[0],"DIS-GLO")) { /* disable glo satellite */
        *q++=0x12;
        *q++=0x02;
        *q++=narg>1?(unsigned char)atoi(args[1]):1; /* slot */
        *q++=0x01;
    }
    else if (!strcmp(args[0],"RAW-RATE")) {
        *q++=0xF4;
        *q++=narg>1?(unsigned char)atoi(args[1]):10; /* interval (100ms) */
    }
#if 0
    else if (!strcmp(args[0],"CFG-SMOOTH")) {
        *q++=ID_XD7SMOOTH;
        *q++=0x03;
        *q++=0x01;
        *q++=0x00;
    }
#endif
    else return 0;
    
    *q++=0x10;
    *q++=0x03;
    
    trace(3,"gen_nvs: msg=\n"); traceb(3,buff,(int)(q-buff));
    
    return (int)(q-buff);
}
